"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

 Asset Catalog View Service Methods - Search for assets, retrieve their properties, lineage and related glossary
 information.

 This file is in active development...

"""

import asyncio

from loguru import logger

from pyegeria.core._server_client import ServerClient
from pyegeria.view.base_report_formats import select_report_format, get_report_spec_match
from pyegeria.models import SearchStringRequestBody, ResultsRequestBody
from pyegeria.view.output_formatter import populate_columns_from_properties, _extract_referenceable_properties, \
    get_required_relationships
from pyegeria.core.utils import body_slimmer, dynamic_catch
from pyegeria.core._globals import max_paging_size
from pyegeria.core._globals import NO_ELEMENTS_FOUND, NO_ASSETS_FOUND
from typing import Any, Optional

class AssetCatalog(ServerClient):
    """
    Client for the Asset Catalog View Service.

    The Asset Catalog View Service provides methods to search for assets,
    retrieve their properties, lineage, and related glossary information.

    Attributes
    ----------
    view_server : str
        The name of the View Server to use.
    platform_url : str
        URL of the server platform to connect to.
    user_id : str
        The identity of the user calling the method.
    user_pwd : str
        The password associated with the user_id. Defaults to None.
    """

    def __init__(self, view_server: str, platform_url: str, user_id: str, user_pwd: Optional[str] = None,
                 token: str = None):
        super().__init__(view_server, platform_url, user_id, user_pwd, token)
        self.view_server = view_server
        self.platform_url = platform_url
        self.user_id = user_id
        self.user_pwd = user_pwd

    #
    # Output helpers
    #

    def _extract_asset_properties(self, element: dict, columns_struct: dict) -> dict:
        col_data = populate_columns_from_properties(element, columns_struct)
        columns_list = col_data.get("formats", {}).get("attributes", [])
        header_props = _extract_referenceable_properties(element)
        for col in columns_list:
            key = col.get("key")
            if key in header_props:
                col["value"] = header_props.get(key)
            elif isinstance(key, str) and key.lower() == "guid":
                col["value"] = header_props.get("GUID")
        # GAP specifics: processStatus, elementTypeName, stepCount
        proc_status = (
            element.get("properties", {}).get("processStatus")
            or element.get("elementProperties", {}).get("processStatus")
            or element.get("processStatus")
        )
        step_count = (
            element.get("properties", {}).get("stepCount")
            or element.get("elementProperties", {}).get("stepCount")
            or element.get("stepCount")
        )
        for col in columns_list:
            key = col.get("key")
            if key in ("process_status", "processStatus"):
                col["value"] = proc_status
            elif key == "stepCount":
                col["value"] = step_count
        col_data = get_required_relationships(element, col_data)
        mermaid_val = element.get("mermaidGraph", "") or ""
        for col in columns_list:
            if col.get("key") == "mermaid":
                col["value"] = mermaid_val
                break
        return col_data

    def _generate_asset_output(self, elements: dict | list[dict], filter_string: Optional[str] = None,
                                            element_type_name: Optional[str] = None, output_format: str = "DICT",
                                            report_spec: dict | str | None = None, **kwargs) -> str | list[dict]:
        return self._generate_formatted_output(
            elements=elements,
            query_string=filter_string,
            entity_type=element_type_name or "Asset",
            output_format=output_format,
            extract_properties_func=self._extract_asset_properties,
            report_spec=report_spec,
            **kwargs
        )

    @dynamic_catch
    async def _async_find_in_asset_domain(
            self,
            search_string: str = "*",
            body: Optional[dict | SearchStringRequestBody] = None,
            starts_with: bool = True,
            ends_with: bool = False,
            ignore_case: bool = False,
            start_from: int = 0,
            page_size: int = 100,
            output_format: str = "JSON",
            report_spec: str | dict = "Referenceable",
            **kwargs
    ) -> list | str:
        """ Retrieve the list of asset metadata elements that contain the search string. Async Version.

        Parameters
        ----------
        search_string : str, default "*"
            Search string to match against - None or '*' indicate match against all assets.
        body : dict | SearchStringRequestBody, optional
            Request body. If provided, overrides other parameters.
        starts_with : bool, default True
            Starts with the supplied string.
        ends_with : bool, default False
            Ends with the supplied string.
        ignore_case : bool, default False
            Ignore case when searching.
        start_from : int, default 0
            Starting index for pagination.
        page_size : int, default 100
            Number of items to return in a single page.
        output_format : str, default "JSON"
            Output format: "MD", "LIST", "FORM", "REPORT", "DICT", "MERMAID" or "JSON".
        report_spec : str | dict, default "Referenceable"
            Report specification for output formatting.
        **kwargs : dict, optional
            Additional parameters supported by the underlying find request:
            
            - anchor_domain : str - Domain to anchor the search
            - metadata_element_type : str - Specific metadata element type
            - metadata_element_subtypes : list[str] - List of metadata element subtypes
            - skip_relationships : list[str] - Relationship types to skip
            - include_only_relationships : list[str] - Only include these relationship types
            - skip_classified_elements : list[str] - Skip elements with these classifications
            - include_only_classified_elements : list[str] - Only include elements with these classifications
            - graph_query_depth : int - Depth of graph traversal (default 3)
            - governance_zone_filter : list[str] - Filter by governance zones
            - as_of_time : str - Historical query time (ISO 8601 format)
            - effective_time : str - Effective time for the query (ISO 8601 format)
            - relationship_page_size : int - Page size for relationships
            - limit_results_by_status : list[str] - Filter by element status
            - sequencing_order : str - Order of results
            - sequencing_property : str - Property to sequence by
            - property_names : list[str] - Specific properties to search

        Returns
        -------
        list | str
            List of assets in the requested format.

        Raises
        ------
        ValidationError
            If the client passes incorrect parameters on the request that don't conform to the data model.
        PyegeriaException
            Issues raised in communicating or server side processing.
        NotAuthorizedException
            The principle specified by the user_id does not have authorization for the requested action.
        """
        url = f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/asset-catalog/assets/in-domain/by-search-string"
        
        # Merge explicit parameters with kwargs
        params = {
            'search_string': search_string,
            'body': body,
            'starts_with': starts_with,
            'ends_with': ends_with,
            'ignore_case': ignore_case,
            'start_from': start_from,
            'page_size': page_size,
            'output_format': output_format,
            'report_spec': report_spec
        }
        params.update(kwargs)
        
        # Filter out None values, but keep search_string even if None (it's required)
        params = {k: v for k, v in params.items() if v is not None or k == 'search_string'}
        
        response = await self._async_find_request(url, _type="Asset", _gen_output=self._generate_asset_output, **params)
        return response

    @dynamic_catch
    def find_in_asset_domain(
            self,
            search_string: str = "*",
            body: Optional[dict | SearchStringRequestBody] = None,
            starts_with: bool = True,
            ends_with: bool = False,
            ignore_case: bool = False,
            start_from: int = 0,
            page_size: int = 100,
            output_format: str = "JSON",
            report_spec: str | dict = "Referenceable",
            **kwargs
    ) -> list | str:
        """ Retrieve the list of asset metadata elements that contain the search string.

        Parameters
        ----------
        search_string : str, default "*"
            Search string to match against - None or '*' indicate match against all assets.
        body : dict | SearchStringRequestBody, optional
            Request body. If provided, overrides other parameters.
        starts_with : bool, default True
            Starts with the supplied string.
        ends_with : bool, default False
            Ends with the supplied string.
        ignore_case : bool, default False
            Ignore case when searching.
        start_from : int, default 0
            Starting index for pagination.
        page_size : int, default 100
            Number of items to return in a single page.
        output_format : str, default "JSON"
            Output format: "MD", "LIST", "FORM", "REPORT", "DICT", "MERMAID" or "JSON".
        report_spec : str | dict, default "Referenceable"
            Report specification for output formatting.
        **kwargs : dict, optional
            Additional parameters supported by the underlying find request:
            
            - anchor_domain : str - Domain to anchor the search
            - metadata_element_type : str - Specific metadata element type
            - metadata_element_subtypes : list[str] - List of metadata element subtypes
            - skip_relationships : list[str] - Relationship types to skip
            - include_only_relationships : list[str] - Only include these relationship types
            - skip_classified_elements : list[str] - Skip elements with these classifications
            - include_only_classified_elements : list[str] - Only include elements with these classifications
            - graph_query_depth : int - Depth of graph traversal (default 3)
            - governance_zone_filter : list[str] - Filter by governance zones
            - as_of_time : str - Historical query time (ISO 8601 format)
            - effective_time : str - Effective time for the query (ISO 8601 format)
            - relationship_page_size : int - Page size for relationships
            - limit_results_by_status : list[str] - Filter by element status
            - sequencing_order : str - Order of results
            - sequencing_property : str - Property to sequence by
            - property_names : list[str] - Specific properties to search

        Returns
        -------
        list | str
            List of assets in the requested format.

        Raises
        ------
        ValidationError
            If the client passes incorrect parameters on the request that don't conform to the data model.
        PyegeriaException
            Issues raised in communicating or server side processing.
        NotAuthorizedException
            The principle specified by the user_id does not have authorization for the requested action.
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self._async_find_in_asset_domain(
                search_string, body, starts_with, ends_with, ignore_case,
                start_from, page_size, output_format, report_spec, **kwargs
            )
        )

    async def _async_get_asset_graph(
        self,
        asset_guid: str,
        start_from: int = 0,
        page_size: int = 0,
        output_format: str = "MERMAID",
        report_spec: str = "Common-Mermaid",
        body: Optional[dict | ResultsRequestBody] = None
    ) -> str | dict:
        """Return all the elements that are anchored to an asset plus relationships between these elements and to
        other elements. Async version.

        Parameters
        ----------
        asset_guid : str
            The unique identity of the asset to get the graph for.
        start_from : int, optional
            The index from which to start fetching. Default is 0.
        page_size : int, optional
            The maximum number of items to fetch. Default is 0 (all).
        output_format : str, default = "MERMAID"
            The desired output format. One of "MD", "LIST", "FORM", "REPORT", "DICT", "MERMAID", or "JSON".
        report_spec : str | dict, optional
            The desired output columns/fields to include. Default is "Common-Mermaid".
        body : dict | ResultsRequestBody, optional
            If provided, the search parameters in the body will supersede other attributes.

        Returns
        -------
        str | dict
            The asset graph in the requested format.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        PyegeriaNotAuthorizedException
            The principle specified by the user_id does not have authorization.
        """

        url = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/asset-catalog/assets/{asset_guid}/"
            f"as-graph"
        )
        response = await self._async_get_results_body_request(url, "Asset", self._generate_asset_output,
                                                        start_from, page_size, output_format, report_spec, body)
        return response

    def get_asset_graph(
        self,
        asset_guid: str,
        start_from: int = 0,
        page_size: int = 0,
        output_format: str = "MERMAID",
        report_spec: str = "Common-Mermaid",
        body: Optional[dict | ResultsRequestBody] = None
    ) -> str | dict:
        """Return all the elements that are anchored to an asset plus relationships between these elements and to
        other elements.

        Parameters
        ----------
        asset_guid : str
            The unique identity of the asset to get the graph for.
        start_from : int, optional
            The index from which to start fetching. Default is 0.
        page_size : int, optional
            The maximum number of items to fetch. Default is 0 (all).
        output_format : str, default = "MERMAID"
            The desired output format. One of "MD", "LIST", "FORM", "REPORT", "DICT", "MERMAID", or "JSON".
        report_spec : str | dict, optional
            The desired output columns/fields to include. Default is "Common-Mermaid".
        body : dict | ResultsRequestBody, optional
            If provided, the search parameters in the body will supersede other attributes.

        Returns
        -------
        str | dict
            The asset graph in the requested format.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        PyegeriaNotAuthorizedException
            The principle specified by the user_id does not have authorization.
        """

        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_get_asset_graph(asset_guid, start_from, page_size, output_format,
                                        report_spec, body)
        )
        return response

    def get_asset_mermaid_graph(
        self,
        asset_guid: str,
        start_from: int = 0,
        page_size: int = 0,
    ) -> str:
        """Return the asset graph as mermaid markdown string.
         Parameters
         ----------
         asset_guid : str
             The unique identity of the asset to get the graph for.

         start_from : int, optional
             The index from which to start fetching the engine actions. Default is 0.

         page_size : int, optional
             The maximum number of engine actions to fetch in a single request. Default is `max_paging_size`.

         Returns
         -------
        str
             A mermaid string representing the asset graph.

         Raises
         ------
         PyegeriaException
             One of the pyegeria exceptions will be raised if there are issues in communications, message format, or
             Egeria errors.
         PyegeriaNotAuthorizedException
             The principle specified by the user_id does not have authorization for the requested action

        """

        asset_graph = self.get_asset_graph(asset_guid, start_from, page_size)
        return asset_graph[0]

    async def _async_get_asset_lineage_graph(
        self,
        asset_guid: str,
        effective_time: Optional[str] = None,
        as_of_time: Optional[str] = None,
        relationship_types: [str] = None,
        limit_to_isc_q_name: Optional[str] = None,
        hilight_isc_q_name: Optional[str] = None,
        all_anchors: bool = False,
        start_from: int = 0,
        page_size: int =0,
        output_format: str = "DICT",
        report_spec: str = "Common-Mermaid",

    ) -> str | dict:
        """Return the asset lineage including a mermaid markdown string. Async version.

        Parameters
        ----------
        asset_guid : str
            The unique identity of the asset to get the graph for.
        effective_time : str, optional
            Effective time to query on. If not specified, the current time is used.
        as_of_time : str, optional
            As-of time to query on. If not specified, the current time is used.
        relationship_types : list[str], optional
            Relationship types to include in the lineage graph. If not specified, all relationship types are included.
        limit_to_isc_q_name : str, optional
            If specified, filters results to only include information supply chains with the given qualified name.
        hilight_isc_q_name : str, optional
            If specified, highlights the information supply chain with the given qualified name.
        all_anchors : bool, optional
            Whether to include all anchors. Default is False.
        start_from : int, optional
            The index from which to start fetching. Default is 0.
        page_size : int, optional
            The maximum number of items to fetch. Default is 0 (all).
        output_format : str, optional
            The desired output format. Default is "DICT".
        report_spec : str, optional
            The desired output columns/fields to include. Default is "Common-Mermaid".

        Returns
        -------
        str | dict
            A dictionary of the asset graph that includes a mermaid markdown string.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """

        url = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/asset-catalog/assets/{asset_guid}/"
            f"as-lineage-graph"
        )
        body = {
            "class": "AssetLineageGraphRequestBody",
            "effectiveTime": effective_time,
            "asOfTime": as_of_time,
            "relationshipTypes": relationship_types,
            "limitToISCQualifiedName": limit_to_isc_q_name,
            "highlightISCQualifiedName": hilight_isc_q_name,
            "allAnchors": all_anchors,
            "startFrom": start_from,
            "pageSize": page_size
            }

        response = await self._async_make_request("POST", url, body_slimmer(body))
        element = response.json().get("element",NO_ASSETS_FOUND)
        if type(element) is str:
            logger.info(NO_ELEMENTS_FOUND)
            return NO_ELEMENTS_FOUND

        if output_format != 'JSON':  # return a simplified markdown representation
            logger.info(f"Found elements, output format: {output_format} and report_spec: {report_spec}")
            return self._generate_asset_output(element, None, "Asset",
                               output_format, report_spec)
        return element


    def get_asset_lineage_graph(
        self,
        asset_guid: str,
        effective_time: Optional[str] = None,
        as_of_time: Optional[str] = None,
        relationship_types: [str] = None,
        limit_to_isc_q_name: Optional[str] = None,
        hilight_isc_q_name: Optional[str] = None,
        all_anchors: bool = False,
        start_from: int = 0,
        page_size: int = 0,
        output_format: str = "DICT",
        report_spec: str = "Common-Mermaid",
        ) -> str | dict:
        """Return the asset lineage including a mermaid markdown string.

        Parameters
        ----------
        asset_guid : str
            The unique identity of the asset to get the graph for.
        effective_time : str, optional
            Effective time to query on. If not specified, the current time is used.
        as_of_time : str, optional
            As-of time to query on. If not specified, the current time is used.
        relationship_types : list[str], optional
            Relationship types to include in the lineage graph. If not specified, all relationship types are included.
        limit_to_isc_q_name : str, optional
            If specified, filters results to only include information supply chains with the given qualified name.
        hilight_isc_q_name : str, optional
            If specified, highlights the information supply chain with the given qualified name.
        all_anchors : bool, optional
            Whether to include all anchors. Default is False.
        start_from : int, optional
            The index from which to start fetching. Default is 0.
        page_size : int, optional
            The maximum number of items to fetch. Default is 0 (all).
        output_format : str, optional
            The desired output format. Default is "DICT".
        report_spec : str, optional
            The desired output columns/fields to include. Default is "Common-Mermaid".

        Returns
        -------
        str | dict
            A dictionary of the asset graph that includes a mermaid markdown string.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """

        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_get_asset_lineage_graph(asset_guid, effective_time, as_of_time, relationship_types,
                                                limit_to_isc_q_name, hilight_isc_q_name, all_anchors,
                                                start_from, page_size, output_format, report_spec)
        )
        return response

    def get_asset_lineage_mermaid_graph(
        self,
        asset_guid: str,
        effective_time: Optional[str] = None,
        as_of_time: Optional[str] = None,
        relationship_types: [str] = None,
        limit_to_isc_q_name: Optional[str] = None,
        hilight_isc_q_name: Optional[str] = None,
        start_from: int = 0,
        page_size: int = max_paging_size,
        ) -> str:
        """Return the asset lineage including a mermaid markdown string. Async Version.
         Parameters
         ----------
         asset_guid : str
             The unique identity of the asset to get the graph for.
        effective_time: str, default is None
            Effective time to query on. If not specified, the current time is used.
        as_of_time: str = None
            as_of_time to query on. If not specified, the current time is used.
        relationship_types: [str], default is None,
            relationship types to include in the lineage graph. If not specified, all relationship types are included.
        limit_to_isc_q_name: Optional[str] = None,
            if specified, filters results to only include information supply chains with the given name.
        hilight_isc_q_name: Optional[str] = None,
            if specified, highlights the information supply chain with the given name.

         start_from : int, optional
             The index from which to start fetching the engine actions. Default is 0.

         page_size : int, optional
             The maximum number of elements to fetch in a single request.
             Default is `max_paging_size`.

        Returns
         -------
        str
             A mermaid string representing the lineage.

         Raises:
         ------
         PyegeriaInvalidParameterException
         PyegeriaAPIException
         PyegeriaUnauthorizedException

    """

        asset_graph = self.get_asset_lineage_graph(asset_guid, effective_time,
                                                   as_of_time, relationship_types,
                                                   limit_to_isc_q_name, hilight_isc_q_name,
                                                   start_from, page_size, "JSON")
        return asset_graph.get("mermaidGraph")


    async def _async_get_assets_by_metadata_collection_id(
        self,
        metadata_collection_id: str,
        type_name: Optional[str] = None,
        effective_time: Optional[str] = None,
        start_from: int = 0,
        page_size: int = 0,
        output_format: str = "JSON",
        report_spec: str = "Referenceable",
    ) -> str | list:
        """Return a list of assets from the requested metadata collection. Async version.

        Parameters
        ----------
        metadata_collection_id : str
            The unique identity of the metadata collection to return assets from.
        type_name : str, optional
            An asset type to filter on. If not specified, all assets in the collection are returned.
        effective_time : str, optional
            The effective time to filter on. If not specified, the current time is used.
        start_from : int, optional
            The index from which to start fetching. Default is 0.
        page_size : int, optional
            The maximum number of assets to fetch in a single request. Default is 0 (all).
        output_format : str, optional
            The desired output format. Default is "JSON".
        report_spec : str, optional
            The desired output columns/fields to include. Default is "Referenceable".

        Returns
        -------
        list | str
            A list of assets or a string if an error occurs or no assets are found.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """

        url = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/asset-catalog/assets/by-metadata-collection-id/"
            f"{metadata_collection_id}"
        )

        body = {"filter": type_name,
                "effectiveTime": effective_time,
                "startFrom": start_from,
                "pageSize": page_size,
                }
        body_s = body_slimmer(body)
        response = await self._async_make_request("POST", url, body_s)
        elements =  response.json().get("elements", "NO_ASSETS_FOUND")
        if type(elements) is str:
            logger.info(NO_ELEMENTS_FOUND)
            return NO_ELEMENTS_FOUND

        if output_format != 'JSON':  # return a simplified markdown representation
            logger.info(f"Found elements, output format: {output_format} and report_spec: {report_spec}")
            return self._generate_asset_output(elements, None, "Asset",
                                               output_format, report_spec)
        return elements

    def get_assets_by_metadata_collection_id(
        self,
        metadata_collection_id: str,
        type_name: Optional[str] = None,
        effective_time: Optional[str] = None,
        start_from: int = 0,
        page_size: int = max_paging_size,
        output_format: str = "JSON",
        report_spec: str = "Referenceable",
    ) -> str | list:
        """Return a list of assets from the requested metadata collection.

        Parameters
        ----------
        metadata_collection_id : str
            The unique identity of the metadata collection to return assets from.
        type_name : str, optional
            An asset type to filter on. If not specified, all assets in the collection are returned.
        effective_time : str, optional
            The effective time to filter on. If not specified, the current time is used.
        start_from : int, optional
            The index from which to start fetching. Default is 0.
        page_size : int, optional
            The maximum number of assets to fetch in a single request. Default is `max_paging_size`.
        output_format : str, optional
            The desired output format. Default is "JSON".
        report_spec : str, optional
            The desired output columns/fields to include. Default is "Referenceable".

        Returns
        -------
        list | str
            A list of assets or a string if an error occurs or no assets are found.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """

        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_get_assets_by_metadata_collection_id(
                metadata_collection_id,
                type_name,
                effective_time,
                start_from,
                page_size,
                output_format,
                report_spec,
            )
        )
        return response




    async def _async_get_asset_types(self) -> list | str:
        """Return the list of asset types supported by the Asset Catalog View Service. Async version.

        Returns
        -------
        list | str
            A list of asset types or NO_ASSETS_FOUND.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """

        url = f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/asset-catalog/assets/types"

        response = await self._async_make_request("GET", url)

        return response.json().get("types", "NO_ASSETS_FOUND")

    def get_asset_types(self) -> list | str:
        """Return the list of asset types supported by the Asset Catalog View Service.

        Returns
        -------
        list | str
            A list of asset types or NO_ASSETS_FOUND.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """

        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(self._async_get_asset_types())
        return response


if __name__ == "__main__":
    print("Main-asset-catalog")



"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

This module tests the DataEngineer class and methods

A running Egeria environment is needed to run these tests.
"""
import asyncio
import time

from rich import print
from rich.console import Console
from pyegeria.omvs.data_engineer import DataEngineer
from pyegeria.core.logging_configuration import config_logging, init_logging
from pyegeria.core._exceptions import (
    PyegeriaInvalidParameterException,
    PyegeriaConnectionException,
    PyegeriaClientException,
    PyegeriaAPIException,
    PyegeriaUnknownException,
    print_basic_exception,
    print_validation_error, )
from pydantic import ValidationError

disable_ssl_warnings = True
console = Console(width=250)

config_logging()
init_logging(True)

VIEW_SERVER = "qs-view-server"
PLATFORM_URL = "https://localhost:9443"
USER_ID = "peterprofile"
USER_PWD = "secret"

class TestDataEngineer:
    good_platform1_url = PLATFORM_URL
    good_user_1 = USER_ID
    good_view_server_1 = VIEW_SERVER

    def test_find_tabular_data_sets(self):
        try:
            de_client = DataEngineer(self.good_view_server_1, self.good_platform1_url, user_id=self.good_user_1, user_pwd='secret')
            de_client.create_egeria_bearer_token(self.good_user_1, "secret")

            start_time = time.perf_counter()
            response = de_client.find_tabular_data_sets(search_string="Attributes", output_format="DICT",report_spec="Referenceable")
            duration = time.perf_counter() - start_time
            
            print(f"response type is {type(response)}, count = {len(response)}")
            print(f"\n\tDuration was {duration} seconds")
            console.print_json(data=response)
            assert response is not None
            assert isinstance(response, list) or isinstance(response, dict) or response == "No elements found"

        except (PyegeriaInvalidParameterException, PyegeriaConnectionException,
                PyegeriaClientException, PyegeriaAPIException, PyegeriaUnknownException) as e:
            print_basic_exception(e)
            assert False, f"Request failed: {e}"
        except ValidationError as e:
            print_validation_error(e)
            assert False, "Validation error"
        finally:
            de_client.close_session()

    def test_get_tabular_data_set(self):
        try:
            de_client = DataEngineer(self.good_view_server_1, self.good_platform1_url, user_id=self.good_user_1, user_pwd='secret')
            de_client.create_egeria_bearer_token(self.good_user_1, "secret")

            # We need a GUID. Let's try to find one first.
            find_response = de_client.find_tabular_data_sets(search_string="Types")
            
            if isinstance(find_response, list) and len(find_response) > 0:
                guid = find_response[0].get('elementHeader', {}).get('guid')

                if guid:
                    start_time = time.perf_counter()
                    response = de_client.get_tabular_data_set(guid)
                    duration = time.perf_counter() - start_time
                    if isinstance(response, dict|list):
                        console.print(f"==>Found {response.get('recordCount',"")} records")
                    print(f"get response type is {type(response)}")
                    print(f"\n\tget Duration was {duration} seconds")

                    assert response is not None
                else:
                    print("No GUID found to test get_tabular_data_set")
            else:
                print("No TabularDataSets found to test get_tabular_data_set")
            
            de_client.close_session()
        except Exception as e:
            print(f"get test failed: {e}")

    def test_get_tabular_data_set_csv(self):
        try:
            de_client = DataEngineer(self.good_view_server_1, self.good_platform1_url, user_id=self.good_user_1,
                                     user_pwd='secret')
            de_client.create_egeria_bearer_token(self.good_user_1, "secret")

            # We need a GUID. Let's try to find one first.
            find_response = de_client.find_tabular_data_sets(search_string="*")

            if isinstance(find_response, list) and len(find_response) > 0:
                guid = find_response[0].get('elementHeader', {}).get('guid')

                if guid:
                    start_time = time.perf_counter()
                    response = de_client.get_tabular_data_set(guid, output_format="RICH-TABLE")
                    duration = time.perf_counter() - start_time
                    if isinstance(response, dict):
                        console.print(f"==>Found {response.get('recordCount', "")} records")
                        console.print(response)
                    print(f"get response type is {type(response)}")
                    print(f"\n\tget Duration was {duration} seconds")

                    assert response is not None
                else:
                    print("No GUID found to test get_tabular_data_set")
            else:
                print("No TabularDataSets found to test get_tabular_data_set")

            de_client.close_session()
        except Exception as e:
            print(f"get test failed: {e}")

    def test_async_find_tabular_data_sets(self):
        async def run_test():
            try:
                de_client = DataEngineer(self.good_view_server_1, self.good_platform1_url, user_id=self.good_user_1, user_pwd='secret')
                de_client.create_egeria_bearer_token(self.good_user_1, "secret")

                start_time = time.perf_counter()
                response = await de_client._async_find_tabular_data_sets(search_string="*")
                duration = time.perf_counter() - start_time
                
                print(f"Async response type is {type(response)}")
                print(f"\n\tAsync Duration was {duration} seconds")
                
                assert response is not None
                de_client.close_session()
            except Exception as e:
                print(f"Async test failed: {e}")
                assert False

        asyncio.run(run_test())

    def test_async_get_tabular_data_set(self):
        async def run_test():
            try:
                de_client = DataEngineer(self.good_view_server_1, self.good_platform1_url, user_id=self.good_user_1, user_pwd='secret')
                de_client.create_egeria_bearer_token(self.good_user_1, "secret")

                # We need a GUID. Let's try to find one first.
                find_response = await de_client._async_find_tabular_data_sets(search_string="*")
                
                if find_response and len(find_response) > 0:
                    # Depending on the output format, find_response might be a list of dicts
                    # If it's a list, take the first one.
                    if isinstance(find_response, list):
                        guid = find_response[0].get('elementHeader', {}).get('guid')
                    elif isinstance(find_response, dict):
                        # Might be a single element or a different structure
                        guid = find_response.get('elementHeader', {}).get('guid')
                    else:
                        guid = None
                    
                    if guid:
                        start_time = time.perf_counter()
                        response = await de_client._async_get_tabular_data_set(guid)
                        duration = time.perf_counter() - start_time
                        
                        print(f"Async get response type is {type(response)}")
                        print(f"\n\tAsync get Duration was {duration} seconds")

                        assert response is not None
                    else:
                        print("No GUID found to test get_tabular_data_set")
                else:
                    print("No TabularDataSets found to test get_tabular_data_set")
                
                de_client.close_session()
            except Exception as e:
                print(f"Async get test failed: {e}")
                # We don't necessarily want to fail if no data exists, but the call itself should work if data exists
                # assert False

        asyncio.run(run_test())



"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

This module provides access to the Digital Business OMVS module.

The Digital Business OMVS provides APIs for managing business capabilities and their
relationships to digital resources.
"""

import asyncio
from typing import Optional

from pyegeria.omvs.collection_manager import CollectionManager
from pyegeria.core._server_client import ServerClient
from pyegeria.models import (
    NewElementRequestBody,
    UpdateElementRequestBody,
    DeleteElementRequestBody,
    NewRelationshipRequestBody,
    DeleteRelationshipRequestBody,
    NewClassificationRequestBody,
    DeleteClassificationRequestBody
)

from pyegeria.core.utils import dynamic_catch, body_slimmer


class DigitalBusiness(CollectionManager):
    """
    Manage business capabilities, their lifecycle, and their digital support relationships.

    This client provides methods to create, update, and manage business capabilities,
    link business capabilities with their dependencies and digital support elements,
    as well as manage business significance classifications.

    Attributes
    ----------
    view_server : str
        The name of the View Server to connect to.
    platform_url : str
        URL of the server platform to connect to.
    user_id : str
        The identity of the user calling the method - this sets a default optionally
        used by the methods when the user doesn't pass the user_id on a method call.
    user_pwd : str, optional
        The password associated with the user_id. Defaults to None.
    token : str, optional
        An optional bearer token for authentication.

    Methods
    -------
    create_business_capability(body)
        Create a new business capability collection.
    update_business_capability(business_capability_guid, body)
        Update the properties of a business capability.
    delete_business_capability(business_capability_guid, body, cascade)
        Delete a business capability.
    get_business_capability_by_guid(business_capability_guid, body, output_format, report_spec)
        Return the properties of a specific business capability by GUID.
    get_business_capabilities_by_name(filter_string, body, start_from, page_size, output_format, report_spec)
        Returns the list of business capabilities with a particular name.
    find_business_capabilities(search_string, starts_with, ends_with, ignore_case, start_from, page_size, output_format, report_spec, body)
        Returns the list of business capabilities matching the search string.
    link_business_capability_dependency(business_capability_guid, supporting_capability_guid, body)
        Link dependent business capabilities.
    detach_business_capability_dependency(business_capability_guid, supporting_capability_guid, body)
        Detach dependent business capabilities.
    link_digital_support(business_capability_guid, element_guid, body)
        Attach a business capability to an element that provides digital support.
    detach_digital_support(business_capability_guid, element_guid, body)
        Detach a business capability from digital support.
    set_business_significant(element_guid, body)
        Classify an element as business significant.
    clear_business_significance(element_guid, body)
        Remove the business significant classification.
    """

    def __init__(
        self,
        view_server: str = None,
        platform_url: str = None,
        user_id: str = None,
        user_pwd: Optional[str] = None,
        token: Optional[str] = None,
    ):
        ServerClient.__init__(self, view_server, platform_url, user_id, user_pwd, token)
        self.view_server = self.server_name
        self.platform_url = self.platform_url
        self.user_id = self.user_id
        self.user_pwd = self.user_pwd
        self.digital_business_command_root: str = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/digital-business"
        )

    def _prepare_body(self, body: Optional[dict | NewElementRequestBody | UpdateElementRequestBody |
                                           DeleteElementRequestBody | NewRelationshipRequestBody |
                                           DeleteRelationshipRequestBody | NewClassificationRequestBody |
                                           DeleteClassificationRequestBody]) -> dict:
        """Convert Pydantic models to dict and slim the body."""
        if body is None:
            return {}
        if isinstance(body, dict):
            return body_slimmer(body)
        # It's a Pydantic model
        return body_slimmer(body.model_dump(mode='json', by_alias=True, exclude_none=True))

    #
    # Business Capability Lifecycle Management
    #

    @dynamic_catch
    async def _async_create_business_capability(
        self,
        body: Optional[dict | NewElementRequestBody],
    ) -> str:
        """Create a new business capability collection. Async version.

        Parameters
        ----------
        body : dict | NewElementRequestBody
            Request body containing business capability properties.

        Returns
        -------
        str
            The GUID of the created business capability.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        ValidationError
            If the body does not conform to NewElementRequestBody.
        PyegeriaNotAuthorizedException
            If the user is not authorized for the requested action.

        Notes
        -----
            JSON Structure looks like:
            {
              "class" : "NewElementRequestBody",
              "typeName": "BusinessCapability",
              "isOwnAnchor" : true,
              "properties": {
                "class" : "BusinessCapabilityProperties",
                "qualifiedName": "BusinessCapability::Add capability name here",
                "displayName" : "Capability name",
                "description" : "Add description of capability here",
                "identifier" : "Add capability identifier here",
                "businessCapabilityType" : "Add type here",
                "additionalProperties": {
                  "property1Name" : "property1Value"
                }
              },
              "effectiveTime" : "timestamp",
              "forLineage" : false,
              "forDuplicateProcessing" : false,
            }
        """
        url = f"{self.digital_business_command_root}/collections"
        return await self._async_create_element_body_request(url, ["BusinessCapabilityProperties"], body)

    def create_business_capability(
        self,
        body: Optional[dict | NewElementRequestBody],
    ) -> str:
        """Create a new business capability collection. Sync version.

        Parameters
        ----------
        body : dict | NewElementRequestBody
            Request body containing business capability properties.

        Returns
        -------
        str
            The GUID of the created business capability.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        body = self._prepare_body(body)
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self._async_create_business_capability(body))

    @dynamic_catch
    async def _async_update_business_capability(
        self,
        business_capability_guid: str,
        body: Optional[dict | UpdateElementRequestBody] = None,
    ) -> None:
        """Update the properties of a business capability. Async version.

        Parameters
        ----------
        business_capability_guid : str
            The GUID of the business capability to update.
        body : dict | UpdateElementRequestBody, optional
            Request body containing updated properties.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        ValidationError
            If the body does not conform to UpdateElementRequestBody.
        PyegeriaNotAuthorizedException
            If the user is not authorized for the requested action.
        """
        url = f"{self.digital_business_command_root}/collections/{business_capability_guid}/update"
        await self._async_update_element_body_request(url, ["BusinessCapabilityProperties"], body)

    def update_business_capability(
        self,
        business_capability_guid: str,
        body: Optional[dict | UpdateElementRequestBody] = None,
    ) -> None:
        """Update the properties of a business capability. Sync version.

        Parameters
        ----------
        business_capability_guid : str
            The GUID of the business capability to update.
        body : dict | UpdateElementRequestBody, optional
            Request body containing updated properties.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        business_capability_guid = str(business_capability_guid)
        body = self._prepare_body(body)
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_update_business_capability(business_capability_guid, body)
        )

    @dynamic_catch
    async def _async_delete_business_capability(
        self,
        business_capability_guid: str,
        body: Optional[dict | DeleteElementRequestBody] = None,
        cascade: bool = False,
    ) -> None:
        """Delete a business capability. Async version.

        Parameters
        ----------
        business_capability_guid : str
            The GUID of the business capability to delete.
        body : dict | DeleteElementRequestBody, optional
            Request body for deletion.
        cascade : bool, optional
            Whether to cascade the delete. Defaults to False.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        ValidationError
            If the body does not conform to DeleteElementRequestBody.
        PyegeriaNotAuthorizedException
            If the user is not authorized for the requested action.
        """
        url = f"{self.digital_business_command_root}/collections/{business_capability_guid}/delete"
        await self._async_delete_element_request(url, body, cascade)

    def delete_business_capability(
        self,
        business_capability_guid: str,
        body: Optional[dict | DeleteElementRequestBody] = None,
        cascade: bool = False,
    ) -> None:
        """Delete a business capability. Sync version.

        Parameters
        ----------
        business_capability_guid : str
            The GUID of the business capability to delete.
        body : dict | DeleteElementRequestBody, optional
            Request body for deletion.
        cascade : bool, optional
            Whether to cascade the delete. Defaults to False.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        business_capability_guid = str(business_capability_guid)
        body = self._prepare_body(body)
        cascade = bool(cascade)
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_delete_business_capability(business_capability_guid, body, cascade)
        )

    @dynamic_catch
    async def _async_get_business_capability_by_guid(
        self,
        business_capability_guid: str,
        body: Optional[dict] = None,
        output_format: str = "JSON",
        report_spec: Optional[str | dict] = None,
    ) -> dict | str:
        """Return the properties of a specific business capability by GUID. Async version.

        Parameters
        ----------
        business_capability_guid : str
            The GUID of the business capability to retrieve.
        body : dict, optional
            Request body (typically empty for retrieval).
        output_format : str, optional
            Format for output. Defaults to "JSON".
        report_spec : str | dict, optional
            Report specification for formatting.

        Returns
        -------
        dict | str
            The business capability properties.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        PyegeriaNotFoundException
            If the business capability is not found.
        PyegeriaNotAuthorizedException
            If the user is not authorized for the requested action.
        """
        url = f"{self.digital_business_command_root}/collections/{business_capability_guid}/retrieve"
        response = await self._async_get_guid_request(
            url,
            _type="BusinessCapability",
            _gen_output=self._generate_collection_output,
            output_format=output_format,
            report_spec=report_spec,
            body=body,
        )
        return response

    def get_business_capability_by_guid(
        self,
        business_capability_guid: str,
        body: Optional[dict] = None,
        output_format: str = "JSON",
        report_spec: Optional[str | dict] = None,
    ) -> dict | str:
        """Return the properties of a specific business capability by GUID. Sync version.

        Parameters
        ----------
        business_capability_guid : str
            The GUID of the business capability to retrieve.
        body : dict, optional
            Request body (typically empty for retrieval).
        output_format : str, optional
            Format for output. Defaults to "JSON".
        report_spec : str | dict, optional
            Report specification for formatting.

        Returns
        -------
        dict | str
            The business capability properties.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self._async_get_business_capability_by_guid(
                business_capability_guid, body, output_format, report_spec
            )
        )

    @dynamic_catch
    async def _async_get_business_capabilities_by_name(
        self,
        filter_string: Optional[str] = None,
        classification_names: Optional[list[str]] = None,
        body: Optional[dict] = None,
        start_from: int = 0,
        page_size: int = 0,
        output_format: str = "JSON",
        report_spec: Optional[str | dict] = None,
    ) -> list | str:
        """Returns the list of business capabilities with a particular name. Async version.

        Parameters
        ----------
        filter_string : str, optional
            Filter string to match against capability names.
        classification_names : list[str], optional
            List of classification names to filter by.
        body : dict, optional
            Request body for additional filtering.
        start_from : int, optional
            Starting index for pagination. Defaults to 0.
        page_size : int, optional
            Number of results per page. Defaults to 0 (no limit).
        output_format : str, optional
            Format for output. Defaults to "JSON".
        report_spec : str | dict, optional
            Report specification for formatting.

        Returns
        -------
        list | str
            List of business capabilities matching the criteria.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        url = f"{self.digital_business_command_root}/collections/by-name"
        response = await self._async_get_name_request(url, _type="BusinessCapability",
                                                      _gen_output=self._generate_collection_output,
                                                      filter_string=filter_string,
                                                      classification_names=classification_names, start_from=start_from,
                                                      page_size=page_size, output_format=output_format,
                                                      report_spec=report_spec, body=body)
        return response

    def get_business_capabilities_by_name(
        self,
        filter_string: Optional[str] = None,
        classification_names: Optional[list[str]] = None,
        body: Optional[dict] = None,
        start_from: int = 0,
        page_size: int = 0,
        output_format: str = "JSON",
        report_spec: Optional[str | dict] = None,
    ) -> list | str:
        """Returns the list of business capabilities with a particular name. Sync version.

        Parameters
        ----------
        filter_string : str, optional
            Filter string to match against capability names.
        classification_names : list[str], optional
            List of classification names to filter by.
        body : dict, optional
            Request body for additional filtering.
        start_from : int, optional
            Starting index for pagination. Defaults to 0.
        page_size : int, optional
            Number of results per page. Defaults to 0 (no limit).
        output_format : str, optional
            Format for output. Defaults to "JSON".
        report_spec : str | dict, optional
            Report specification for formatting.

        Returns
        -------
        list | str
            List of business capabilities matching the criteria.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self._async_get_business_capabilities_by_name(
                filter_string, classification_names, body, start_from, page_size, output_format, report_spec
            )
        )

    @dynamic_catch
    async def _async_find_business_capabilities(
        self,
        search_string: str = "*",
        body: Optional[dict] = None,
        starts_with: bool = False,
        ends_with: bool = False,
        ignore_case: bool = True,
        start_from: int = 0,
        page_size: int = 0,
        output_format: str = "JSON",
        report_spec: Optional[str | dict] = None,
        **kwargs
    ) -> list | str:
        """Returns the list of business capabilities matching the search string. Async version.

        Parameters
        ----------
        search_string : str, optional
            Search string to match. Defaults to "*" (all).
        starts_with : bool, optional
            Whether to match from the start. Defaults to False.
        ends_with : bool, optional
            Whether to match at the end. Defaults to False.
        ignore_case : bool, optional
            Whether to ignore case. Defaults to True.
        anchor_domain : str, optional
            Domain to anchor the search.
        metadata_element_type : str, optional
            Type of metadata element to search for. Defaults to "BusinessCapability".
        metadata_element_subtype : list[str], optional
            Subtypes of metadata element.
        skip_relationships : list[str], optional
            Relationships to skip in the graph.
        include_only_relationships : list[str], optional
            Only include these relationships.
        skip_classified_elements : list[str], optional
            Skip elements with these classifications.
        include_only_classified_elements : list[str], optional
            Only include elements with these classifications.
        graph_query_depth : int, optional
            Depth of graph query. Defaults to 0.
        governance_zone_filter : list[str], optional
            Filter by governance zones.
        as_of_time : str, optional
            Historical time for the query.
        effective_time : str, optional
            Effective time for the query.
        relationship_page_size : int, optional
            Page size for relationships. Defaults to 0.
        limit_results_by_status : list[str], optional
            Limit results by status values.
        sequencing_order : str, optional
            Order for sequencing results.
        sequencing_property : str, optional
            Property to sequence by.
        start_from : int, optional
            Starting index for pagination. Defaults to 0.
        page_size : int, optional
            Number of results per page. Defaults to 0 (no limit).
        output_format : str, optional
            Format for output. Defaults to "JSON".
        report_spec : str | dict, optional
            Report specification for formatting.
        property_names: list[str], optional
            The names of properties to search for.
        body : dict, optional
            Request body for additional parameters.

        Returns
        -------
        list | str
            List of business capabilities matching the search criteria.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        url = f"{self.digital_business_command_root}/collections/by-search-string"
        
        # Merge explicit parameters with kwargs
        params = {
            'search_string': search_string,
            'body': body,
            'starts_with': starts_with,
            'ends_with': ends_with,
            'ignore_case': ignore_case,
            'start_from': start_from,
            'page_size': page_size,
            'output_format': output_format,
            'report_spec': report_spec
        }
        params.update(kwargs)
        
        # Filter out None values, but keep search_string even if None (it's required)
        params = {k: v for k, v in params.items() if v is not None or k == 'search_string'}
        
        response = await self._async_find_request(url, _type="BusinessCapability",
                                                  _gen_output=self._generate_collection_output, **params)
        return response

    def find_business_capabilities(
        self,
        search_string: str = "*",
        body: Optional[dict] = None,
        starts_with: bool = False,
        ends_with: bool = False,
        ignore_case: bool = True,
        start_from: int = 0,
        page_size: int = 0,
        output_format: str = "JSON",
        report_spec: Optional[str | dict] = None,
        **kwargs
    ) -> list | str:
        """Returns the list of business capabilities matching the search string. Sync version.

        Parameters
        ----------
        search_string : str, optional
            Search string to match. Defaults to "*" (all).
        starts_with : bool, optional
            Whether to match from the start. Defaults to False.
        ends_with : bool, optional
            Whether to match at the end. Defaults to False.
        ignore_case : bool, optional
            Whether to ignore case. Defaults to True.
        anchor_domain : str, optional
            Domain to anchor the search.
        metadata_element_type : str, optional
            Type of metadata element to search for. Defaults to "BusinessCapability".
        metadata_element_subtype : list[str], optional
            Subtypes of metadata element.
        skip_relationships : list[str], optional
            Relationships to skip in the graph.
        include_only_relationships : list[str], optional
            Only include these relationships.
        skip_classified_elements : list[str], optional
            Skip elements with these classifications.
        include_only_classified_elements : list[str], optional
            Only include elements with these classifications.
        graph_query_depth : int, optional
            Depth of graph query. Defaults to 0.
        governance_zone_filter : list[str], optional
            Filter by governance zones.
        as_of_time : str, optional
            Historical time for the query.
        effective_time : str, optional
            Effective time for the query.
        relationship_page_size : int, optional
            Page size for relationships. Defaults to 0.
        limit_results_by_status : list[str], optional
            Limit results by status values.
        sequencing_order : str, optional
            Order for sequencing results.
        sequencing_property : str, optional
            Property to sequence by.
        start_from : int, optional
            Starting index for pagination. Defaults to 0.
        page_size : int, optional
            Number of results per page. Defaults to 0 (no limit).
        output_format : str, optional
            Format for output. Defaults to "JSON".
        report_spec : str | dict, optional
            Report specification for formatting.
        property_names: list[str], optional
            The names of properties to search for.
        body : dict, optional
            Request body for additional parameters.

        Returns
        -------
        list | str
            List of business capabilities matching the search criteria.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self._async_find_business_capabilities(
                search_string=search_string,
                body=body,
                starts_with=starts_with,
                ends_with=ends_with,
                ignore_case=ignore_case,
                start_from=start_from,
                page_size=page_size,
                output_format=output_format,
                report_spec=report_spec,
                **kwargs
            )
        )

    #
    # Business Capability Dependency Management
    #

    @dynamic_catch
    async def _async_link_business_capability_dependency(
        self,
        business_capability_guid: str,
        supporting_capability_guid: str,
        body: Optional[dict | NewRelationshipRequestBody] = None,
    ) -> None:
        """Link dependent business capabilities. Async version.

        Parameters
        ----------
        business_capability_guid : str
            The GUID of the business capability that depends on another.
        supporting_capability_guid : str
            The GUID of the supporting business capability.
        body : dict | NewRelationshipRequestBody, optional
            Request body containing relationship properties.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        ValidationError
            If the body does not conform to NewRelationshipRequestBody.
        PyegeriaNotAuthorizedException
            If the user is not authorized for the requested action.
        """
        url = (
            f"{self.digital_business_command_root}/business-capabilities/"
            f"{business_capability_guid}/dependencies/{supporting_capability_guid}/attach"
        )
        await self._async_new_relationship_request(url=url, prop=['BusinessCapabilityDependencyProperties'],body=body)

    def link_business_capability_dependency(
        self,
        business_capability_guid: str,
        supporting_capability_guid: str,
        body: Optional[dict | NewRelationshipRequestBody] = None,
    ) -> None:
        """Link dependent business capabilities. Sync version.

        Parameters
        ----------
        business_capability_guid : str
            The GUID of the business capability that depends on another.
        supporting_capability_guid : str
            The GUID of the supporting business capability.
        body : dict | NewRelationshipRequestBody, optional
            Request body containing relationship properties.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_link_business_capability_dependency(
                business_capability_guid, supporting_capability_guid, body
            )
        )

    @dynamic_catch
    async def _async_detach_business_capability_dependency(
        self,
        business_capability_guid: str,
        supporting_capability_guid: str,
        body: Optional[dict | DeleteRelationshipRequestBody] = None,
        cascade: bool = False,
    ) -> None:
        """Detach dependent business capabilities. Async version.

        Parameters
        ----------
        business_capability_guid : str
            The GUID of the business capability.
        supporting_capability_guid : str
            The GUID of the supporting business capability to detach.
        body : dict | DeleteRelationshipRequestBody, optional
            Request body for deletion.
        cascade : bool, optional, default=False
            Whether to cascade delete dependent elements.
        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.

        """
        url = (
            f"{self.digital_business_command_root}/business-capabilities/"
            f"{business_capability_guid}/dependencies/{supporting_capability_guid}/detach"
        )
        await self._async_delete_relationship_request(url=url, body=body, cascade_delete=cascade)

    def detach_business_capability_dependency(
        self,
        business_capability_guid: str,
        supporting_capability_guid: str,
        body: Optional[dict | DeleteRelationshipRequestBody] = None,
        cascade: bool = False,
    ) -> None:
        """Detach dependent business capabilities. Sync version.

        Parameters
        ----------
        business_capability_guid : str
            The GUID of the business capability.
        supporting_capability_guid : str
            The GUID of the supporting business capability to detach.
        body : dict | DeleteRelationshipRequestBody, optional
            Request body for deletion.
        cascade : bool, optional, default=False
            Whether to cascade delete dependent elements.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_detach_business_capability_dependency(business_capability_guid, supporting_capability_guid,
                                                              body, cascade)
        )

    #
    # Digital Support Management
    #

    @dynamic_catch
    async def _async_link_digital_support(
        self,
        business_capability_guid: str,
        element_guid: str,
        body: Optional[dict | NewRelationshipRequestBody] = None,
    ) -> None:
        """Attach a business capability to an element that provides digital support. Async version.

        Parameters
        ----------
        business_capability_guid : str
            The GUID of the business capability.
        element_guid : str
            The GUID of the element providing digital support.
        body : dict | NewRelationshipRequestBody, optional
            Request body containing relationship properties.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        url = (
            f"{self.digital_business_command_root}/business-capabilities/"
            f"{business_capability_guid}/digital-support/{element_guid}/attach"
        )
        await self._async_new_relationship_request(url=url, prop=['DigitalSupportProperties'],body=body)

    def link_digital_support(
        self,
        business_capability_guid: str,
        element_guid: str,
        body: Optional[dict | NewRelationshipRequestBody] = None,
    ) -> None:
        """Attach a business capability to an element that provides digital support. Sync version.

        Parameters
        ----------
        business_capability_guid : str
            The GUID of the business capability.
        element_guid : str
            The GUID of the element providing digital support.
        body : dict | NewRelationshipRequestBody, optional
            Request body containing relationship properties.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_link_digital_support(business_capability_guid, element_guid, body)
        )

    @dynamic_catch
    async def _async_detach_digital_support(
        self,
        business_capability_guid: str,
        element_guid: str,
        body: Optional[dict | DeleteRelationshipRequestBody] = None,
        cascade: bool = False,
    ) -> None:
        """Detach a business capability from digital support. Async version.

        Parameters
        ----------
        business_capability_guid : str
            The GUID of the business capability.
        element_guid : str
            The GUID of the digital support element to detach.
        body : dict | DeleteRelationshipRequestBody, optional
            Request body for deletion.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.

        Args:
            cascade ():
        """
        url = (
            f"{self.digital_business_command_root}/business-capabilities/"
            f"{business_capability_guid}/digital-support/{element_guid}/detach"
        )
        await self._async_delete_relationship_request(url, self._prepare_body(body), cascade)

    def detach_digital_support(
        self,
        business_capability_guid: str,
        element_guid: str,
        body: Optional[dict | DeleteRelationshipRequestBody] = None,
        cascade: bool = False,
    ) -> None:
        """Detach a business capability from digital support. Sync version.

        Parameters
        ----------
        business_capability_guid : str
            The GUID of the business capability.
        element_guid : str
            The GUID of the digital support element to detach.
        body : dict | DeleteRelationshipRequestBody, optional
            Request body for deletion.
        cascade: bool, default False
            If true, will perform a cascade delete.
        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.


        Args:
            cascade ():
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_detach_digital_support(business_capability_guid, element_guid, body, cascade)
        )

    #
    # Business Significance Classification
    #

    @dynamic_catch
    async def _async_set_business_significant(
        self,
        element_guid: str,
        body: Optional[dict | NewClassificationRequestBody] = None,
    ) -> None:
        """Classify an element as business significant. Async version.

        Parameters
        ----------
        element_guid : str
            The GUID of the element to classify.
        body : dict | NewClassificationRequestBody, optional
            Request body containing classification properties.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        url = f"{self.digital_business_command_root}/elements/{element_guid}/business-significant"
        await self._async_new_classification_request(url=url, prop = ['BusinessSignificantProperties'],body=body)

    def set_business_significant(
        self,
        element_guid: str,
        body: Optional[dict | NewClassificationRequestBody] = None,
    ) -> None:
        """Classify an element as business significant. Sync version.

        Parameters
        ----------
        element_guid : str
            The GUID of the element to classify.
        body : dict | NewClassificationRequestBody, optional
            Request body containing classification properties.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._async_set_business_significant(element_guid, body))

    @dynamic_catch
    async def _async_clear_business_significance(
        self,
        element_guid: str,
        body: Optional[dict | DeleteClassificationRequestBody] = None,
    ) -> None:
        """Remove the business significant classification. Async version.

        Parameters
        ----------
        element_guid : str
            The GUID of the element to declassify.
        body : dict | DeleteClassificationRequestBody, optional
            Request body for deletion.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        url = f"{self.digital_business_command_root}/elements/{element_guid}/business-significant/remove"
        await self._async_delete_classification_request(url=url,body=body,)

    def clear_business_significance(
        self,
        element_guid: str,
        body: Optional[dict | DeleteClassificationRequestBody] = None,
    ) -> None:
        """Remove the business significant classification. Sync version.

        Parameters
        ----------
        element_guid : str
            The GUID of the element to declassify.
        body : dict | DeleteClassificationRequestBody, optional
            Request body for deletion.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._async_clear_business_significance(element_guid, body))